/* vim: set sts=4 ts=4 sw=4 et nojs: */
using System;
using System.IO;
using System.Text;

/**
 *  ntfs_I30_analyze.cs:
 *      $I30 (directory) index analysis program.
 */

namespace RobbieUtils
{
    public class I30ArgumentException : ApplicationException
    {
        public I30ArgumentException() : base()
        {
        }

        public I30ArgumentException(string msg) : base(msg)
        {
        }

        public I30ArgumentException(string msg, Exception inner) : base(msg, inner)
        {
        }
    }

    public class NtfsI30Analyzer
    {
        public static int Main(string[] args)
        {
            return (new NtfsI30Analyzer()).DoThings(args);
        }

        public int DoThings(string[] args)
        {
            try {
                bool additionalArgNeeded = false;
                string previousSwitch = "";
                string filePath = "";
                string offsetStr = "";
                bool isRootHeader = false;
                int offset;
                foreach (string arg in args) {
                    if (additionalArgNeeded) {
                        if (previousSwitch == "-o") {
                            offsetStr = arg;
                        } else {
                            throw new InvalidOperationException();
                        }
                        additionalArgNeeded = false;
                    } else {
                        if (arg.StartsWith("-")) {
                            if (arg == "-o") {
                                previousSwitch = arg;
                                additionalArgNeeded = true;
                            } else if (arg == "-r") {
                                isRootHeader = true;
                            } else {
                                Console.WriteLine("{0} is not a valid switch", arg);
                                throw new I30ArgumentException();
                            }
                        } else {
                            filePath = arg;
                        }
                    }
                }
                if (additionalArgNeeded) {
                    Console.WriteLine("You must provide the argument to switch {0}.",
                        previousSwitch);
                    ShowHelp();
                    return 2;
                }
                if (filePath == "") {
                    Console.WriteLine("You must specify the file path.");
                    ShowHelp();
                    return 2;
                }
                if (offsetStr != "") {
                    try {
                        offset = Int32.Parse(offsetStr);
                    } catch (OverflowException ex) {
                        if (ex == null) { /* dummy use */ }
                        Console.WriteLine("Offset {0} is too large or too small.", offsetStr);
                        ShowHelp();
                        return 2;
                    } catch (FormatException ex) {
                        if (ex == null) { /* dummy use */ }
                        Console.WriteLine("Offset {0} is not an integer.", offsetStr);
                        ShowHelp();
                        return 2;
                    }
                } else {
                    offset = 0;
                }
                Analyze(filePath, offset, isRootHeader);
                return 0;
            } catch (I30ArgumentException ex) {
                /* Message already sent */
                if (ex == null) { /* dummy use */ }
                ShowHelp();
                return 2;
            } catch (Exception ex) {
                Console.WriteLine(ex.ToString());
                Console.WriteLine(ex.StackTrace);
                return 1;
            }
        }

        public void ShowHelp()
        {
            Console.WriteLine("Analyzes a cluster dump file for NTFS $I30 (directory indexes.");
            Console.WriteLine("Usage: ntfs_I30_analyze [-o <offset>] [-r] <dumpfile.dat>");
            Console.WriteLine("-o <offset> - specifies the offset in dump file from where analysis should");
            Console.WriteLine("    begin");
            Console.WriteLine("-r - parses the header for $INDEX_ROOT; by default, the program assumes that");
            Console.WriteLine("    there is a standard index header at the beginning of the dump file");
        }

        public void Analyze(string filePath, int offset, bool isRootHeader)
        {
            FileStream reader = null;
            NtfsIndexHeader header;
            NtfsIndexEntry entry;
            long beginPos;
            bool isLast;

            try {
                /**
                 *  NOTE: Seek() and Read() are permissive.
                 *  Prototypes:
                 *  Seek(int offset, SeekOrigin origin);
                 *      SeekOrigin.Begin
                 *      SeekOrigin.Current
                 *      SeekOrigin.End
                 *  Read(byte[] buffer, int offsetInBuffer, int length);
                 */
                reader = new FileStream(filePath, FileMode.Open, FileAccess.Read,
                    FileShare.Read);
                Console.WriteLine("Seek offset: 0x{0:X4}", offset);
                reader.Seek(offset, SeekOrigin.Begin);
                beginPos = reader.Position;
                if (isRootHeader) {
                    header = ParseRootHeader(reader);
                } else {
                    header = ParseHeader(reader);
                }
                while ((reader.Position - beginPos - header.OffsetExcluded)
                       < header.IndexEntrySize) {
                    entry = ParseEntry(reader, out isLast);
                }
                Console.WriteLine("Parsing complete. Now at {0:X4}", reader.Position);
            } finally {
                if (reader != null) {
                    reader.Close();
                }
            }
        }

        public NtfsIndexHeader ParseHeader(FileStream reader)
        {
            NtfsIndexHeader result = new NtfsIndexHeader();
            long beginPos;

            beginPos = reader.Position;
            result.MagicNumber = Encoding.UTF8.GetString(ReadBytesFromStream(reader, 4));
            Console.WriteLine(result.MagicNumber);
            result.UpdSeqOffset = BitConverter.ToUInt16(ReadBytesFromStream(reader, 2), 0);
            result.UsnAndArraySize = BitConverter.ToUInt16(ReadBytesFromStream(reader, 2), 0);
            result.Lsn = BitConverter.ToUInt64(ReadBytesFromStream(reader, 8), 0);
            Console.WriteLine("LSN: {0:X16}", result.Lsn);
            result.IndexBufferVcn = BitConverter.ToUInt64(ReadBytesFromStream(reader, 8), 0);
            Console.WriteLine("VCN: {0:X16}", result.IndexBufferVcn);
            result.IndexEntryOffset = (int)BitConverter.ToUInt32(ReadBytesFromStream(reader, 4), 0);
            Console.WriteLine("Index entry offset: {0} (0x{0:X2})", result.IndexEntryOffset);
            result.IndexEntrySize = (int)BitConverter.ToUInt32(ReadBytesFromStream(reader, 4), 0);
            Console.WriteLine("Index entry size: {0} (0x{0:X4})", result.IndexEntrySize);
            result.IndexEntryAllocSize = (int)BitConverter.ToUInt32(ReadBytesFromStream(reader, 4), 0);
            Console.WriteLine("Index entry allocated size: {0} (0x{0:X4})", result.IndexEntryAllocSize);
            result.LeafNodeFlag = BitConverter.ToBoolean(ReadBytesFromStream(reader, 1), 0);
            Console.WriteLine("Leaf node: {0}", result.LeafNodeFlag);
            ReadBytesFromStream(reader, 3); // Padding
            result.Usn = BitConverter.ToUInt16(ReadBytesFromStream(reader, 2), 0);
            Console.WriteLine("USN: {0}", result.Usn);
            result.UpdSeqArray = ReadBytesFromStream(reader, result.UsnAndArraySize * 2 - 2);
            Console.WriteLine("Update sequence array: {0}", FormatBytes(result.UpdSeqArray));
            ReadBytesFromStream(reader, result.IndexEntryOffset + 0x18 - (int)(reader.Position - beginPos));
            return result;
        }

        public NtfsIndexHeader ParseRootHeader(FileStream reader)
        {
            NtfsIndexHeader result = new NtfsIndexHeader();
            long beginPos;

            beginPos = reader.Position;
            result.IsRootHeader = true;
            result.IndexEntryOffset = (int)BitConverter.ToUInt32(ReadBytesFromStream(reader, 4), 0);
            Console.WriteLine("Index entry offset: {0} (0x{0:X2})", result.IndexEntryOffset);
            result.IndexEntrySize = (int)BitConverter.ToUInt32(ReadBytesFromStream(reader, 4), 0);
            Console.WriteLine("Index entry size: {0} (0x{0:X4})", result.IndexEntrySize);
            result.IndexEntryAllocSize = (int)BitConverter.ToUInt32(ReadBytesFromStream(reader, 4), 0);
            Console.WriteLine("Index entry allocated size: {0} (0x{0:X4})", result.IndexEntryAllocSize);
            result.LargeIndexFlag = BitConverter.ToBoolean(ReadBytesFromStream(reader, 1), 0);
            Console.WriteLine("Large index: {0}", result.LargeIndexFlag);
            ReadBytesFromStream(reader, 3); // Padding
            return result;
        }

        public NtfsIndexEntry ParseEntry(FileStream reader, out bool isLast)
        {
            NtfsIndexEntry result = new NtfsIndexEntry();
            long beginPos;

            beginPos = reader.Position;
            result.FileRef = BitConverter.ToUInt64(ReadBytesFromStream(reader, 8), 0);
            Console.WriteLine("File reference to: {0:X12}, sequence number: {1:X4}",
                result.FileRef & 0xFFFFFFFFFFFF, result.FileRef >> 48);
            result.EntrySize = BitConverter.ToUInt16(ReadBytesFromStream(reader, 2), 0);
            result.FileNameOffset = BitConverter.ToUInt16(ReadBytesFromStream(reader, 2), 0);
            result.IndexFlags = BitConverter.ToUInt16(ReadBytesFromStream(reader, 2), 0);
            isLast = (((IndexFlags)result.IndexFlags & IndexFlags.IsLastEntry) != 0);
            Console.WriteLine("Index flags: {0:X4} ({1})", result.IndexFlags,
                (IndexFlags)result.IndexFlags);
            ReadBytesFromStream(reader, 2); // Padding
            if (isLast) {
                /* All remaining data will be garbage */
                if (((IndexFlags)result.IndexFlags & IndexFlags.HasSubNode) != 0) {
                    result.SubNodeVcn = BitConverter.ToUInt64(ReadBytesFromStream(reader, 8), 0);
                    Console.WriteLine("Sub node VCN: {0:X16}", result.SubNodeVcn);
                }
                Console.WriteLine("0x{0:X4}", reader.Position);
                return result;
            }
            result.ParentFileRef = BitConverter.ToUInt64(ReadBytesFromStream(reader, 8), 0);
            Console.WriteLine("Parent file record: {0:X12}, sequence number: {1:X4}",
                result.ParentFileRef & 0xFFFFFFFFFFFF, result.ParentFileRef >> 48);
            result.CreationTime = BitConverter.ToUInt64(ReadBytesFromStream(reader, 8), 0);
            Console.WriteLine("Creation time: {0}", FormatFileTime(result.CreationTime));
            result.ModificationTime = BitConverter.ToUInt64(ReadBytesFromStream(reader, 8), 0);
            Console.WriteLine("Modification time: {0}", FormatFileTime(result.ModificationTime));
            result.FileRecordModTime = BitConverter.ToUInt64(ReadBytesFromStream(reader, 8), 0);
            Console.WriteLine("File record modified time: {0}", FormatFileTime(result.FileRecordModTime));
            result.AccessTime = BitConverter.ToUInt64(ReadBytesFromStream(reader, 8), 0);
            Console.WriteLine("Access time: {0}", FormatFileTime(result.AccessTime));
            result.AllocatedSize = BitConverter.ToUInt64(ReadBytesFromStream(reader, 8), 0);
            result.RealSize = BitConverter.ToUInt64(ReadBytesFromStream(reader, 8), 0);
            result.FileFlags = BitConverter.ToUInt32(ReadBytesFromStream(reader, 4), 0);
            ReadBytesFromStream(reader, 4); // Padding
            result.FileNameLength = ReadBytesFromStream(reader, 1)[0];
            Console.WriteLine("File name length: {0}", result.FileNameLength);
            result.FileNameNamespace = ReadBytesFromStream(reader, 1)[0];
            Console.WriteLine("File name namespace: {0}", FormatFileNameNamespace(
                result.FileNameNamespace));
            Console.WriteLine("Let's see the fallback object of Encoding.Unicode.DecoderFallback: {0}",
                Encoding.Unicode.DecoderFallback.GetType());
            result.FileName = Encoding.Unicode.GetString(ReadBytesFromStream(reader, result.FileNameLength * 2));
            Console.WriteLine("File name: {0}", result.FileName);
            Console.WriteLine("0x{0:X4}", reader.Position);
            Console.WriteLine((8 - (int)(reader.Position - beginPos) % 8) % 8);
            ReadBytesFromStream(reader, (8 - (int)(reader.Position - beginPos) % 8) % 8); // Padding
            Console.WriteLine("0x{0:X4}", reader.Position);
            if (((IndexFlags)result.IndexFlags & IndexFlags.HasSubNode) != 0) {
                result.SubNodeVcn = BitConverter.ToUInt64(ReadBytesFromStream(reader, 8), 0);
                Console.WriteLine("Sub node VCN: {0:X16}", result.SubNodeVcn);
            }
            return result;
        }

        public byte[] ReadBytesFromStream(FileStream reader, int length)
        {
            byte[] result = new byte[length];
            int actualLength;

            actualLength = reader.Read(result, 0, length);
            if (actualLength != length) {
                Array.Resize(ref result, actualLength);
            }
            return result;
        }

        public string FormatBytes(byte[] bytes)
        {
            StringBuilder sb = new StringBuilder();
            int i;

            for (i = 0; i < bytes.Length; i++) {
                byte b = bytes[i];
                if (i > 0) {
                    if (i % 16 == 0) {
                        sb.Append(Environment.NewLine);
                    } else if (i % 8 == 0) {
                        sb.Append(" - ");
                    } else {
                        sb.Append(" ");
                    }
                }
                sb.Append(b.ToString("X2"));
            }
            return sb.ToString();
        }

        public string FormatFileTime(ulong fileTime)
        {
            string result = DateTime.FromFileTime((long)fileTime)
                .ToString("yyyy-MM-dd HH:mm:ss.ffffff %K");
            return result;
        }

        public string FormatFileNameNamespace(int namspc)
        {
            if (namspc == 0) {
                return "POSIX";
            } else if (namspc == 1) {
                return "Win32";
            } else if (namspc == 2) {
                return "DOS";
            } else if (namspc == 3) {
                return "Win32 & DOS";
            } else {
                return "Unknown namespace";
            }
        }
    }

    [Flags]
    public enum IndexFlags
    {
        HasSubNode = 0x01,
        IsLastEntry = 0x02,
    }

    public class NtfsIndexHeader
    {
        private bool _isRootHeader;
        private string _magicNumber;
        private int _updSeqOffset;
        private int _usnAndArraySize;
        private ulong _lsn; /* log sequence number */
        private ulong _indexBufferVcn;
        private int _indexEntryOffset;
        private int _indexEntrySize;
        private int _indexEntryAllocSize;
        private bool _leafNodeFlag;
        private bool _largeIndexFlag;
        private int _usn;
        private byte[] _updSeqArray;

        public NtfsIndexHeader()
        {
        }

        public bool IsRootHeader
        {
            get { return _isRootHeader; }
            set { _isRootHeader = value; }
        }

        public int OffsetExcluded
        {
            get
            { 
                if (_isRootHeader) { 
                    return 0x00;
                } else {
                    return 0x18;
                }
            }
        }

        public string MagicNumber
        {
            get { return _magicNumber; }
            set { _magicNumber = value; }
        }

        public int UpdSeqOffset
        {
            get { return _updSeqOffset; }
            set { _updSeqOffset = value; }
        }

        public int UsnAndArraySize
        {
            get { return _usnAndArraySize; }
            set { _usnAndArraySize = value; }
        }

        public ulong Lsn
        {
            get { return _lsn; }
            set { _lsn = value; }
        }

        public ulong IndexBufferVcn
        {
            get { return _indexBufferVcn; }
            set { _indexBufferVcn = value; }
        }

        public int IndexEntryOffset
        {
            get { return _indexEntryOffset; }
            set { _indexEntryOffset = value; }
        }

        public int IndexEntrySize
        {
            get { return _indexEntrySize; }
            set { _indexEntrySize = value; }
        }

        public int IndexEntryAllocSize
        {
            get { return _indexEntryAllocSize; }
            set { _indexEntryAllocSize = value; }
        }

        public bool LeafNodeFlag
        {
            get { return _leafNodeFlag; }
            set { _leafNodeFlag = value; }
        }

        public bool LargeIndexFlag
        {
            get { return _largeIndexFlag; }
            set { _largeIndexFlag = value; }
        }

        public int Usn
        {
            get { return _usn; }
            set { _usn = value; }
        }

        public byte[] UpdSeqArray
        {
            get { return _updSeqArray; }
            set { _updSeqArray = value; }
        }
    }

    public class NtfsIndexEntry
    {
        private ulong _fileRef;
        private int _entrySize;
        private int _fileNameOffset;
        private ushort _indexFlags;
        private ulong _parentFileRef;
        private ulong _creationTime;
        private ulong _modificationTime;
        private ulong _fileRecordModTime;
        private ulong _accessTime;
        private ulong _allocatedSize;
        private ulong _realSize;
        private uint _fileFlags;
        private int _fileNameLength;
        private int _fileNameNamespace;
        private string _fileName;
        private ulong _subNodeVcn;

        public NtfsIndexEntry()
        {
        }

        public ulong FileRef
        {
            get { return _fileRef; }
            set { _fileRef = value; }
        }

        public int EntrySize
        {
            get { return _entrySize; }
            set { _entrySize = value; }
        }

        public int FileNameOffset
        {
            get { return _fileNameOffset; }
            set { _fileNameOffset = value; }
        }

        public ushort IndexFlags
        {
            get { return _indexFlags; }
            set { _indexFlags = value; }
        }

        public ulong ParentFileRef
        {
            get { return _parentFileRef; }
            set { _parentFileRef = value; }
        }

        public ulong CreationTime
        {
            get { return _creationTime; }
            set { _creationTime = value; }
        }

        public ulong ModificationTime
        {
            get { return _modificationTime; }
            set { _modificationTime = value; }
        }

        public ulong FileRecordModTime
        {
            get { return _fileRecordModTime; }
            set { _fileRecordModTime = value; }
        }

        public ulong AccessTime
        {
            get { return _accessTime; }
            set { _accessTime = value; }
        }

        public ulong AllocatedSize
        {
            get { return _allocatedSize; }
            set { _allocatedSize = value; }
        }

        public ulong RealSize
        {
            get { return _realSize; }
            set { _realSize = value; }
        }

        public uint FileFlags
        {
            get { return _fileFlags; }
            set { _fileFlags = value; }
        }

        public int FileNameLength
        {
            get { return _fileNameLength; }
            set { _fileNameLength = value; }
        }

        public int FileNameNamespace
        {
            get { return _fileNameNamespace; }
            set { _fileNameNamespace = value; }
        }

        public string FileName
        {
            get { return _fileName; }
            set { _fileName = value; }
        }

        public ulong SubNodeVcn
        {
            get { return _subNodeVcn; }
            set { _subNodeVcn = value; }
        }
    }
}
